
#import "RCWOLittleHelper.h"
#import <Foundation/NSBundle.h>
#import <Foundation/NSPathUtilities.h>
#import <AppKit/NSTableView.h>
#import <AppKit/NSTextView.h>
#import <AppKit/NSTextStorage.h>
#import <stdlib.h>
#import <objc/objc-runtime.h>

static void InstallFixes();
static IMP RCOverrideInstanceMethodWithSelector(Class aClass, SEL oldSelector, SEL newSelector);

@implementation RCWOLittleHelper
/* 
 * Install the mouse event stuff upon the InputManager being loaded. Do it in
 * -init as well just in case +load fails.
 */
+ (void)load
{
    InstallFixes();
}
- init
{
    InstallFixes();
    return self;
}
@end

@interface RCWOXHackTableView : NSTableView
@end
@interface RCWOXHackTextView : NSTextView
@end

static IMP origSetLastTextContainer = NULL;

static void InstallFixes()
{
    static BOOL installed = NO;
    NSString *appName;
    
    if (installed)
        return;

    installed = YES;
    appName = [[[[NSBundle mainBundle] bundlePath] lastPathComponent] stringByDeletingPathExtension];

    if ([appName isEqual:@"EOModeler"] ||
        //[appName isEqual:@"ProjectBuilderWO"] ||  // doesn't seem to help anything
        [appName isEqual:@"WebObjects Builder"])
    {
        [RCWOXHackTableView poseAsClass:[NSTableView class]];
    }

    if ([appName isEqual:@"WebObjects Builder"])
    {
        [RCWOXHackTextView poseAsClass:[NSTextView class]];

        /* We can't pose for HTMLAttachmentCell since we can't be sure its framework (HTMLEditing) is loaded */
        origSetLastTextContainer = RCOverrideInstanceMethodWithSelector(NSClassFromString(@"HTMLAttachmentCell"),
                                                                        @selector(setLastTextContainer:),
                                                                        @selector(__rc_htmlAttachmentCell_setLastTextContainer:));
    }
}

@implementation RCWOXHackTableView

- (void)reloadData
{
    int editRow = [self editedRow];
    int editCol = [self editedColumn];
    int selRow  = [self selectedRow];

    [super reloadData];

    /* 
     * On MacOS X, NSTableView's -reloadData method now ends any editing that
     * was going on. This makes tabbing through table cells impossible, as the
     * act of tabbing triggers this method, which then ends editing. So, we
     * check to see if editing was in progress, and if so, restore the editing.
     * 
     * The real problem in EOModeler is in EOTableViewAssociation's
     * subjectChanged method (which calls reloadData but doesn't restore the
     * editing), and for WOBuilder in (at least) HTMLCustomMarkerInspector's
     * initializeUserInterface method, which does the same thing. EOModeler has
     * direct support for bundles, but I don't know of anything easy for
     * WebObjects Builder, so we just do a single input manager to deal with
     * both apps, and hacking this one method fixes things in both places.
     */
    if (editRow >= 0 && editCol >= 0 &&    // we were editing
        [self editedRow] == -1 &&          // we're no longer editing
        [self editedColumn] == -1 &&
        [self selectedRow] == selRow &&    // selectedRow hasn't changed
        editRow < [self numberOfRows] &&   // it's safe to still edit
        editCol < [self numberOfColumns])
    {
        [self editColumn:editCol row:editRow withEvent:nil select:YES];
    }
}

@end


@interface NSTextStorage (ApplePrivate)
- (BOOL)_isEditing;
@end

@implementation RCWOXHackTextView
/* 
 * This crash happens when you create a table, then inside one of the cells
 * create another table, then put the cursor inside one of the inner table's
 * cells, then hit undo. An exception is raised saying that layout was done
 * while the text storage was editing. This method is in the backtrace, so we
 * override to avoid layout if the text storage is editing. Presumably the area
 * will be re-rendered (and laid out) later on.
 */
- (void)setNeedsDisplayInRect:(NSRect)rect avoidAdditionalLayout:(BOOL)flag
{
    if ([[self textStorage] _isEditing])
        flag = YES;
    [super setNeedsDisplayInRect:rect avoidAdditionalLayout:flag];
}

@end

@interface NSObject (Apple)
- (id)lastTextContainer;
@end

@implementation NSObject (RCHTMLAttachmentCellFix)
/* 
 * This crash happens when you create a table, select the whole table (not just
 * a row), and click the create table button again (creating a new table around
 * the first one), then hit undo. The exception that gets raised is in
 * setLastTextContainer, saying there's already a text container set. So, we
 * override to nil out the existing ivar to avoid the exception. This works,
 * though sometimes I see areas of HTML "disappear", but going to raw HTML mode
 * and back forces everything to redisplay right.
 * 
 * Since we are an input manager loaded into all apps, we can't count on the
 * HTMLAttachmentCell class being availble, so we can't pose for it (because we
 * can't subclass it, which posing needs). So, we have to play runtime games,
 * and insert this method into the HTMLAttachmentCell class at runtime.
 */   
- (void)__rc_htmlAttachmentCell_setLastTextContainer:(id)container
{
    id currContainer = [self lastTextContainer];
    
    if (container && currContainer && (currContainer != container))
    {
        // nil out the ivar directly.  Calling origSetLastTextContainer() with nil isn't enough.
        object_setInstanceVariable(self, "_lastTextContainer", nil);
        //[currContainer autorelease]; //causes problems... hopefully was never retained to begin with
    }

    // call the real method
    origSetLastTextContainer(self, @selector(setLastTextContainer:), container);
}
@end



/** Functions to play runtime games. **/
static void _RCAddMethod(IMP methodImp, Class class, const char *methodTypes, SEL selector)
{
    struct objc_method_list *newMethodList;

    newMethodList = (struct objc_method_list *)NSZoneMalloc(NSDefaultMallocZone(), sizeof(struct objc_method_list));

    newMethodList->method_count = 1;
    newMethodList->method_list[0].method_name = selector;
    newMethodList->method_list[0].method_imp = methodImp;
    newMethodList->method_list[0].method_types = (char *)methodTypes;
    class_addMethods(class, newMethodList);
}

static IMP RCOverrideInstanceMethodWithSelector(Class aClass, SEL oldSelector, SEL newSelector)
{
    struct objc_method *oldMethod, *newMethod;
    IMP                 oldImp = NULL;

    if ((oldMethod = class_getInstanceMethod(aClass, oldSelector))) {
        if (newMethod = class_getInstanceMethod(aClass, newSelector)) {
            oldImp = oldMethod->method_imp;
            _RCAddMethod(newMethod->method_imp, aClass, oldMethod->method_types, oldSelector);
        }
    }

    return oldImp;
}
